#ifndef __FDPSIMULATOR_H__
#define __FDPSIMULATOR_H__

#include "defines.h"

class FDPSimulator
{
public:
    typedef enum ALGORITHM_TYPE
    {
    	CPU,
	GPU_BASIC,
	GPU_USE_FLOAT2,
	//PROPER_REDUCTION,
	GPU_FORCE_CALC,
	GPU_FORCE_CALC_KIN,
	LOCAL_MEMORY,
	DATA_SIZE,
	MATRIX_TRANSPOSE,
	USE_GRID_BLOCKS,
	FORCE_REDUCTION,
	FLOAT4_FORCES,
	MAX_ALGORITHM,	
	MIN_ALGORITHM = CPU,
    } ALGORITHM;

    FDPSimulator(unsigned char** graph, unsigned int num_nodes);
    ~FDPSimulator();

    void init();
    void simulate(ALGORITHM alg = CPU);
    void dump();
    
    const char* get_name(ALGORITHM alg) const;
    float* get_pos_x() { return m_pos_x; }
    float* get_pos_y() { return m_pos_y; }

private:
    void update_force();
    void update_vel();
    void update_pos();
    void calc_force(float this_x, float this_y, float other_x, float other_y,
unsigned char edge_weight, float* force_x, float* force_y);
    float get_kinetic_energy();

    unsigned int m_num_nodes;
    unsigned char** m_graph;

    float* m_pos_x;
    float* m_pos_y;
    float* m_vel_x;
    float* m_vel_y;
    float* m_force_x;
    float* m_force_y;

    float m_damping;
    float m_timestep;
};

#endif
