#include <stdio.h>
#include <cutil.h>
#include <cuda.h>

#include "defines.h"


//This kernel is extra stupid.  Each block consists of 32 threads which just computes the 
//value of this block.
__global__ void basic_fps_kernel ( 	unsigned char* graph_data,
					float* x_pos, 
					float* y_pos, 
					float* x_vel,
					float* y_vel,
					float* final_xpos,
					float* final_ypos,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (index < N)
	{
	
		float force_x = 0.0f;
		float force_y = 0.0f;
		
		//step 1 - compute force on this node
		for (int i = 0; i < N; ++i)
		{
			if (i != index)
			{
				float dist_x = x_pos[i] - x_pos[index];
				float dist_y = y_pos[i] - y_pos[index];
				float dist_squared = dist_x * dist_x + dist_y * dist_y;
				float dist = sqrt(dist_squared);
				
				force_x += -dist_x * REPULSIVE_FORCE_CONSTANT / dist_squared / dist;
				force_y += -dist_y * REPULSIVE_FORCE_CONSTANT / dist_squared / dist;
				
				force_x += ATTRACTIVE_FORCE_CONSTANT * (dist_x - dist_x * 100 / dist) * graph_data[i * N + index];
				force_y += ATTRACTIVE_FORCE_CONSTANT * (dist_y - dist_y * 100 / dist) * graph_data[i * N + index];
			}
		}	
		x_vel[index] += (TIMESTEP * force_x);
		y_vel[index] += (TIMESTEP * force_y);
		
		x_vel[index] *= DAMPING;
		y_vel[index] *= DAMPING;
		
		final_xpos[index] = x_pos[index] + TIMESTEP * x_vel[index];
		final_ypos[index] = y_pos[index] + TIMESTEP * y_vel[index];			
	}
}

//This kernel is extra stupid. We just have a single thread copute the energy.  
//should use the reduction technique from Assn1 .
static __global__ void compute_energy  ( float* xvel,
					float* yvel,
					float*  energy,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	if ( index == 0 )
	{
		*energy = 0.0f;
		for (unsigned int i = 0 ; i < N; ++ i )
		{
			*energy += (xvel[i] * xvel[i] + yvel[i] * yvel[i]);
		}
	}
}

extern "C" void basic_fps_gpu_algorithm
(
	unsigned int num_nodes,
	unsigned char** graph,
	float* x_pos,
	float* y_pos
)
{
	float *d_xpos, *d_ypos, *d_xvel, *d_yvel, *d_final_xpos, *d_final_ypos;
	unsigned char* d_graph_data;
	unsigned char* mapped_graph_data;
	float *h_xvel, *h_yvel;
	float *d_kenetic_energy;
	
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_xpos, sizeof(float) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_ypos, sizeof(float) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_xvel, sizeof(float) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_yvel, sizeof(float) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_final_xpos, sizeof(float) * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_final_ypos, sizeof(float) * num_nodes));	
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_kenetic_energy, sizeof(float)));	
	
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_graph_data, sizeof(unsigned char) * num_nodes * num_nodes));
	
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_xpos, (void*) x_pos, sizeof(float) * num_nodes, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_ypos, (void*) y_pos, sizeof(float) * num_nodes, cudaMemcpyHostToDevice));

	mapped_graph_data = (unsigned char*)malloc(num_nodes * num_nodes * sizeof(unsigned char) );
	h_xvel = (float*)malloc(sizeof(float) * num_nodes);
	h_yvel = (float*)malloc(sizeof(float) * num_nodes);
	for (int i = 0; i < num_nodes; ++i)
	{
		for (int j = 0; j < num_nodes; ++j )
		{
			mapped_graph_data[i * num_nodes + j] = graph[i][j];
		}
		h_xvel[i] = 0.0f;
		h_yvel[i] = 0.0f;
	}	
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_graph_data, (void*) mapped_graph_data, sizeof(unsigned char) * num_nodes* num_nodes, cudaMemcpyHostToDevice));		
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_xvel, (void*) h_xvel, sizeof(float*) *num_nodes, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_yvel, (void*) h_yvel, sizeof(float*) *num_nodes, cudaMemcpyHostToDevice));	

	int counter = 0;
	float kenetic_energy = 0.0f;
	do 
	{
		basic_fps_kernel<<< (num_nodes + 31) / 32, 32>>>(d_graph_data, d_xpos, d_ypos, d_xvel,d_yvel, d_final_xpos, d_final_ypos, num_nodes);						
		cudaThreadSynchronize();
		
		compute_energy<<< 1, 1 >>>(d_xvel, d_yvel, d_kenetic_energy, num_nodes);
		cudaThreadSynchronize();

		CUDA_SAFE_CALL(cudaMemcpy ( (void*) &kenetic_energy, (void*) d_kenetic_energy, sizeof(float), cudaMemcpyDeviceToHost));
		cudaThreadSynchronize();
		
		float *tmp;
		tmp = d_xpos;
		d_xpos = d_final_xpos;
		d_final_xpos = tmp;
		
		tmp = d_ypos;
		d_ypos = d_final_ypos;
		d_final_ypos = tmp;
		
		if (DEBUG) if (counter % 10000 == 1) printf("Energy = %f\n", kenetic_energy);
		counter++;
	} while (kenetic_energy  > STABLE_ENERGY_THRESHOLD * num_nodes);
	
	printf("Total iterations were %d\n", counter);
	printf("Energy = %f\n", kenetic_energy);
	
	// d_pos contains the last result because pointers were swapped
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) x_pos, (void*) d_xpos, sizeof(float) * num_nodes, cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) y_pos, (void*) d_ypos, sizeof(float) * num_nodes, cudaMemcpyDeviceToHost));
	
	free(mapped_graph_data);
	free(h_xvel);
	free(h_yvel);
	cudaFree( d_xvel);
	cudaFree( d_yvel);
	cudaFree( d_xpos);
	cudaFree( d_ypos);
	cudaFree( d_final_xpos);
	cudaFree( d_final_ypos);
	cudaFree( d_graph_data);
	cudaFree( d_kenetic_energy);
}
