#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <set>
#include <vector>

#include <cuda.h>
#include <cutil.h>

#include "defines.h"
#include "FDPSimulator.h"
#include "DOTReaderWriter.h"

using namespace std;

unsigned int htimer;

int main(int argc, char** argv)
{
    unsigned char** graph;
    DOTReaderWriter* dot = 0;
    FDPSimulator* sim = 0;

#if USE_DEFAULT_GRAPH
    const unsigned int NUM_NODES = 8;

    graph = (unsigned char**)malloc(NUM_NODES * sizeof(unsigned char*));
    for (unsigned int i = 0; i < NUM_NODES; ++i)
    {
        graph[i] = (unsigned char*)malloc(NUM_NODES * sizeof(unsigned char));
        memset(graph[i], (unsigned char)0, NUM_NODES * sizeof(unsigned char));
    }
    graph[0][1] = 1; graph[1][0] = 1;
    graph[1][2] = 1; graph[2][1] = 1;
    graph[0][3] = 1; graph[3][0] = 1;
    graph[0][4] = 1; graph[4][0] = 1;
    graph[2][5] = 1; graph[5][2] = 1;
    graph[2][6] = 1; graph[6][2] = 1;
    graph[3][5] = 1; graph[5][3] = 1;
    graph[0][6] = 1; graph[6][0] = 1;
    graph[2][7] = 1; graph[7][2] = 1;

    printf("Using default graph...\n");
    printf("1 -> 2,4,5,7\n");
    printf("2 -> 3\n");
    printf("3 -> 6,7,8\n");
    printf("4 -> 6\n");

    sim = new FDPSimulator(graph, NUM_NODES);
#else
   // parse arguments
    if (argc != 3)
    {
        printf("Usage: %s <input_file> <output_file>\n", argv[0]);
        exit(1);
    }
    dot = new DOTReaderWriter(argv[1], argv[2]);
    char* out_file = argv[2];

    // process input file
    unsigned int num_nodes = dot->read(graph);
    printf("Num nodes: %d\n", num_nodes);

    // check for disconnected graph
    set<int> connected_set;
    set<int> last_set;
    last_set.insert(0);
    while (1)
    {
        set<int> new_set;
        set<int>::iterator it;
        for (it = last_set.begin(); it != last_set.end(); it++)
        {
            int i = *it;
            for (unsigned int j = 0; j < num_nodes; ++j)
            {
                if (graph[i][j] != 0 &&
                    last_set.find(j) == last_set.end() &&
                    connected_set.find(j) == connected_set.end())
                {
                    new_set.insert(j);
                }
            }
            connected_set.insert(i);
        }
        
        if (new_set.size() == 0) break;
        last_set = new_set;
    }
    if (connected_set.size() != num_nodes)
    {
        printf("Disconnected graph!\n");
        exit(1);
    }

    sim = new FDPSimulator(graph, num_nodes);
#endif

    if (DEBUG) printf("\nInitial random placement\n");
    // if (DEBUG) sim->dump();  don't dump before init

    cutCreateTimer(&htimer);
    //for ( int i = FDPSimulator::MIN_ALGORITHM; i < FDPSimulator::MAX_ALGORITHM; ++i) 
    //for ( int i = FDPSimulator::GPU_FORCE_CALC; i < FDPSimulator::MAX_ALGORITHM; ++i) 
    for ( int i = FDPSimulator::MAX_ALGORITHM - 1; i >= FDPSimulator::MIN_ALGORITHM; --i) 
    //int i = FDPSimulator::MATRIX_TRANSPOSE;
    {
        if (i == FDPSimulator::CPU && num_nodes >= 1000) continue;

	// random initial placement
        sim->init();

	// dump initial placement
        char new_file_name[1024];
        sprintf(new_file_name, "%s.init", out_file);
        dot->write(sim->get_pos_x(), sim->get_pos_y());
        rename(out_file, new_file_name);

    	printf("====================================");
    	printf("\n%s\n", sim->get_name(static_cast<FDPSimulator::ALGORITHM>(i)));	

	// run
    	cutResetTimer(htimer);
    	cutStartTimer(htimer);
    	sim->simulate(static_cast<FDPSimulator::ALGORITHM>(i));
    	cutStopTimer(htimer);

    	printf("Processing time was %f\n", cutGetTimerValue(htimer));
	
	//if (DEBUG) sim->dump();

	// dump final placement
        sprintf(new_file_name, "%s.%d", out_file, i);
        dot->write(sim->get_pos_x(), sim->get_pos_y());
        rename(out_file, new_file_name);
    }
    return 0;
}

