
// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
// includes, project
#include <cuda.h>
#include <cutil.h>
#include <limits.h>

#define REDUCTION_THREAD_COUNT 		512
#define REDUCTION_MAX_THREAD_COUNT 	512
#define REDUCTION_BLOCK_COUNT 		32
#define REDUCTION_SECTION_SIZE		(REDUCTION_MAX_THREAD_COUNT / REDUCTION_BLOCK_COUNT)
#define WARP_SIZE			32

static float* d_reduction_intermediate_values;
static float* d_cuda_max;

//Out magic array add function
static __global__ void cuda_final_coalese (float *da, int N, float *energy)
{
	__shared__ float energy_array[REDUCTION_BLOCK_COUNT * REDUCTION_SECTION_SIZE];
	
	energy_array[threadIdx.x] = da[threadIdx.x];
	
	__syncthreads();
	
	// User reduction technique presented in class
	if (threadIdx.x < 256) { energy_array[threadIdx.x] += energy_array[threadIdx.x +  256 ]; } 
	__syncthreads(); 
	if (threadIdx.x < 128) { energy_array[threadIdx.x] += energy_array[threadIdx.x + 128]; }
	__syncthreads(); 
	if (threadIdx.x <  64) { energy_array[threadIdx.x] += energy_array[threadIdx.x + 64]; }
	__syncthreads();
        
	if (threadIdx.x < 32)
	{
		energy_array[threadIdx.x] += energy_array[threadIdx.x+32];
		energy_array[threadIdx.x] += energy_array[threadIdx.x+16];
		energy_array[threadIdx.x] += energy_array[threadIdx.x+8];
		energy_array[threadIdx.x] += energy_array[threadIdx.x+4];
		energy_array[threadIdx.x] += energy_array[threadIdx.x+2];
		energy_array[threadIdx.x] += energy_array[threadIdx.x+1];
	}
	
	if (threadIdx.x == 0) { *energy = energy_array[0]; }
}


//Out magic array reductio function from documentation
static __global__ void cuda_compute_energy (float2 *da, int N, float *cuda_intermediate_result)
{
	__shared__ float local_energy_array[512];
	unsigned int i;
		
	const unsigned int elements_per_thread = (N / (gridDim.x * blockDim.x) ) + 1;
	const unsigned int base_id = (blockIdx.x * blockDim.x * elements_per_thread) + threadIdx.x; 
	const unsigned int last_id_tmp = base_id + (blockDim.x * elements_per_thread);
	const unsigned int last_id  = (last_id_tmp < N  ? last_id_tmp : N );
	
	//instantiate local energy to 0 if we have no iterations to do
	local_energy_array[threadIdx.x] = 0.0f;

	float2 next_velocity;
	float local_energy = 0.0f;
	for ( i = base_id; i < last_id; i += REDUCTION_THREAD_COUNT)
	{
		next_velocity = da[i];
		local_energy += next_velocity.x * next_velocity.x + next_velocity.y * next_velocity.y;
	}
	
	local_energy_array[threadIdx.x] = local_energy;
	
	__syncthreads();

	// Use reduction technique presented in class.
	if (threadIdx.x < 256) { local_energy_array[threadIdx.x] +=local_energy_array[threadIdx.x +256]; }
	__syncthreads(); 
	if (threadIdx.x < 128) { local_energy_array[threadIdx.x] +=local_energy_array[threadIdx.x+128]; } 
	__syncthreads(); 
	if (threadIdx.x <  64) { local_energy_array[threadIdx.x] +=local_energy_array[threadIdx.x+64]; } 
	__syncthreads();
        
	if (threadIdx.x < 32)
	{
	        local_energy_array[threadIdx.x] += local_energy_array[threadIdx.x + 32];  
        	local_energy_array[threadIdx.x] += local_energy_array[threadIdx.x + 16];  
	}

    	//stop once we hit section size.  Push the section size back into memory and use that for 
	//the final coalese.
	if (threadIdx.x < REDUCTION_SECTION_SIZE)
	{
		cuda_intermediate_result[blockIdx.x * REDUCTION_SECTION_SIZE + threadIdx.x]=local_energy_array[threadIdx.x];
	}
}

////////////Setup an array for a specific run
static void prepare_reduction_memory()
{
	cudaMalloc ( (void**) &d_reduction_intermediate_values, sizeof (float) * REDUCTION_MAX_THREAD_COUNT );
	cudaMalloc ( (void**) &d_cuda_max, sizeof(float) );
}

////////////Setup an array for a specific run
static void free_reduction_memory()
{
	cudaFree ( &d_reduction_intermediate_values);
	cudaFree( &d_cuda_max);
}

static void do_reduction(float2* d_vel, int size, float* energy)
{
	cuda_compute_energy <<<REDUCTION_BLOCK_COUNT, REDUCTION_THREAD_COUNT>>>(d_vel, size, d_reduction_intermediate_values);
	cuda_final_coalese <<< 1 , REDUCTION_MAX_THREAD_COUNT >>> ( d_reduction_intermediate_values,REDUCTION_BLOCK_COUNT * REDUCTION_SECTION_SIZE, d_cuda_max);
	cudaMemcpy ((void *) energy, (void *) d_cuda_max, sizeof (int), cudaMemcpyDeviceToHost);		
	//cudaThreadSynchronize (); //shouldn't really be necessary
}
