
// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
// includes, project
#include <cuda.h>
#include <cutil.h>
#include <limits.h>

#define REDUCTION_THREAD_COUNT 		512
#define REDUCTION_MAX_THREAD_COUNT 	512
#define REDUCTION_BLOCK_COUNT 		32
#define REDUCTION_SECTION_SIZE		(REDUCTION_MAX_THREAD_COUNT / REDUCTION_BLOCK_COUNT)
#define WARP_SIZE			32

static float2* d_rforce_tmp;
static float2* d_aforce_tmp;

#define float2add(a,b) { a.x += b.x; a.y += b.y; }

//Out magic array add function
static __global__ void cuda_final_coalese (float2 *d_rforce, float2 *d_aforce, int N, float2 *d_force_per_node)
{
	__shared__ float2 rforce_array[REDUCTION_BLOCK_COUNT * REDUCTION_SECTION_SIZE];
	__shared__ float2 aforce_array[REDUCTION_BLOCK_COUNT * REDUCTION_SECTION_SIZE];
	
	rforce_array[threadIdx.x] = d_rforce[blockIdx.x * N + threadIdx.x];
	aforce_array[threadIdx.x] = d_aforce[blockIdx.x * N + threadIdx.x];
	
	__syncthreads();
	
	// User reduction technique presented in class
	if (threadIdx.x < 256) {
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x +  256 ]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x +  256 ]);
	} 
	__syncthreads(); 
	if (threadIdx.x < 128) {
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x + 128]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x + 128]);
	}
	__syncthreads(); 
	if (threadIdx.x <  64) {
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x + 64]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x + 64]);
	}
	__syncthreads();
        
	if (threadIdx.x < 32)
	{
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+32]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+32]);
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+16]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+16]);
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+8]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+8]);
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+4]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+4]);
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+2]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+2]);
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+1]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+1]);
	}
	
	if (threadIdx.x == 0) {
		float2 my_rforce = rforce_array[0];
		float2 my_aforce = aforce_array[0];
		my_rforce.x *= REPULSIVE_FORCE_CONSTANT;
		my_rforce.y *= REPULSIVE_FORCE_CONSTANT;
		my_aforce.x *= ATTRACTIVE_FORCE_CONSTANT;
		my_aforce.y *= ATTRACTIVE_FORCE_CONSTANT;

		float2add(my_rforce, my_aforce);
		d_force_per_node[blockIdx.x] = my_rforce;
	}
}


//Out magic array reductio function from documentation
static __global__ void cuda_compute_energy (float2 *d_rforce, float2 *d_aforce, int node, int N, float2 *d_rforce_tmp, float2 *d_aforce_tmp)
{
	__shared__ float2 rforce_array[512];
	__shared__ float2 aforce_array[512];
	unsigned int i;
		
	const unsigned int elements_per_thread = (N * N / (gridDim.x * blockDim.x) ) + 1;
	const unsigned int base_id = node*N + (blockIdx.x * blockDim.x * elements_per_thread) + threadIdx.x; 
	const unsigned int last_id_tmp = base_id + (blockDim.x * elements_per_thread);
	const unsigned int last_id  = (last_id_tmp < ((node+1)*N) ? last_id_tmp : ((node+1)*N) );
	
	float2 local_rforce;
	float2 local_aforce;
	local_rforce.x = 0.0f;
	local_rforce.y = 0.0f;
	local_aforce.x = 0.0f;
	local_aforce.y = 0.0f;
	for ( i = base_id; i < last_id; i += REDUCTION_THREAD_COUNT)
	{
		float2add(local_rforce, d_rforce[i]);
		float2add(local_aforce, d_aforce[i]);
	}
	
	rforce_array[threadIdx.x] = local_rforce;
	aforce_array[threadIdx.x] = local_aforce;
	
	__syncthreads();

	// Use reduction technique presented in class.
	if (threadIdx.x < 256) {
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x +256]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x +256]);
	}
	__syncthreads(); 
	if (threadIdx.x < 128) {
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+128]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+128]);
	} 
	__syncthreads(); 
	if (threadIdx.x <  64) {
		float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x+64]);
		float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x+64]);
	}
	__syncthreads();
        
	if (threadIdx.x < 32)
	{
	        float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x + 32]);  
	        float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x + 32]);  
        	float2add(rforce_array[threadIdx.x], rforce_array[threadIdx.x + 16]);  
        	float2add(aforce_array[threadIdx.x], aforce_array[threadIdx.x + 16]);  
	}

    	//stop once we hit section size.  Push the section size back into memory and use that for 
	//the final coalese.
	if (threadIdx.x < REDUCTION_SECTION_SIZE)
	{
		d_rforce_tmp[(node * REDUCTION_BLOCK_COUNT + blockIdx.x) * REDUCTION_SECTION_SIZE + threadIdx.x] = rforce_array[threadIdx.x];
		d_aforce_tmp[(node * REDUCTION_BLOCK_COUNT + blockIdx.x) * REDUCTION_SECTION_SIZE + threadIdx.x] = aforce_array[threadIdx.x];
	}
}

////////////Setup an array for a specific run
static void prepare_reduction_force_memory(int size)
{
	cudaMalloc ( (void**) &d_rforce_tmp, sizeof (float2) * REDUCTION_MAX_THREAD_COUNT * size);
	cudaMalloc ( (void**) &d_aforce_tmp, sizeof (float2) * REDUCTION_MAX_THREAD_COUNT * size);
}

////////////Setup an array for a specific run
static void free_reduction_force_memory()
{
	cudaFree ( &d_rforce_tmp);
	cudaFree ( &d_aforce_tmp);
}

static void do_reduction_force(float2* d_rforce, float2* d_aforce, int size, float2* d_force_per_node)
{
	for (int i = 0; i < size; ++i)
		cuda_compute_energy <<<REDUCTION_BLOCK_COUNT, REDUCTION_THREAD_COUNT>>>(d_rforce, d_aforce, i, size, d_rforce_tmp, d_aforce_tmp);
	cudaThreadSynchronize ();
	cuda_final_coalese <<< size , REDUCTION_MAX_THREAD_COUNT >>> ( d_rforce_tmp, d_aforce_tmp ,REDUCTION_BLOCK_COUNT * REDUCTION_SECTION_SIZE, d_force_per_node);
	cudaThreadSynchronize ();
}

